/* swfoutput.cc
   Implements generation of swf files using the rfxswf lib. The routines
   in this file are called from pdf2swf.

   This file is part of swftools.

   Swftools is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Swftools is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with swftools; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include "swfoutput.h"
#include "spline.h"
extern "C" {
#include "../lib/log.h"
#include "../lib/rfxswf.h"
}

static int fi;
static int flag_protected;
static SWF swf;
static TAG *tag;
static int shapeid = 0;
static SHAPE* shape;
static int fillstyleid;
static int linestyleid;
static int swflastx=0;
static int swflasty=0;
static int lastwasfill = 0;
static char* filename = 0;
static int sizex;
static int sizey;

// matrix multiplication. changes p0
void transform (plotxy*p0,struct swfmatrix*m)
{
    double x,y;
    x = m->m11*p0->x+m->m12*p0->y;
    y = m->m21*p0->x+m->m22*p0->y;
    p0->x = x + m->m13;
    p0->y = y + m->m23;
}

// write a move-to command into the swf
void moveto(plotxy p0)
{
    int rx = (int)(p0.x*20);
    int ry = (int)(p0.y*20);
    if(rx!=swflastx || ry!=swflasty) {
      ShapeSetMove (tag, shape, rx,ry);
    }
    swflastx=rx;
    swflasty=ry;
}

// write a line-to command into the swf
void lineto(plotxy p0)
{
    int rx = ((int)(p0.x*20)-swflastx);
    int ry = ((int)(p0.y*20)-swflasty);
    /* we can't skip this for rx=0,ry=0, those
       are plots */
    ShapeSetLine (tag, shape, rx,ry);
    swflastx+=rx;
    swflasty+=ry;
}

// write a spline-to command into the swf
void splineto(plotxy control,plotxy end)
{
    int cx = ((int)(control.x*20)-swflastx);
    int cy = ((int)(control.y*20)-swflasty);
    swflastx += cx;
    swflasty += cy;
    int ex = ((int)(end.x*20)-swflastx);
    int ey = ((int)(end.y*20)-swflasty);
    swflastx += ex;
    swflasty += ey;
    ShapeSetCurve(tag, shape, cx,cy,ex,ey);
}

/* write a line, given two points and the transformation
   matrix. */
void line(plotxy p0, plotxy p1, struct swfmatrix*m)
{
    transform(&p0,m);
    transform(&p1,m);
    moveto(p0);
    lineto(p1);
}

/* write a cubic (!) spline. This involves calling the approximate()
   function out of spline.cc to convert it to a quadratic spline.  */
void spline(plotxy p0,plotxy p1,plotxy p2,plotxy p3,struct swfmatrix*m)
{
    double d;
    struct qspline q[16];
    int num;
    int t;
    transform(&p0,m);
    transform(&p1,m);
    transform(&p2,m);
    transform(&p3,m);

    num = approximate(p0,p1,p2,p3,q);
    for(t=0;t<num;t++) {
	moveto(q[t].start);
	splineto(q[t].control, q[t].end);
    }
}

/* draw a T1 outline. These are generated by pdf2swf and by t1lib.
   (representing characters) */
void drawpath(T1_OUTLINE*outline, struct swfmatrix*m)
{
    double x=0,y=0;
    double lastx=0,lasty=0;

    while (outline)
    {
	logf("<debug> Pathtype:%s",outline->type == T1_PATHTYPE_MOVE?"MOVE":
						    (outline->type == T1_PATHTYPE_LINE?"LINE"
										      :"BEZIER"));
	logf("<debug> relative coordinates: %08x,%08x", outline->dest.x, outline->dest.y);
	x += (outline->dest.x/(float)0xffff);
	y += (outline->dest.y/(float)0xffff);
	logf("<debug> coordinates: %f,%f", x, y);
	if(outline->type == T1_PATHTYPE_MOVE)
	{
	}
	else if(outline->type == T1_PATHTYPE_LINE) 
	{
	    plotxy p0;
	    plotxy p1;
	    p0.x=lastx;
	    p0.y=lasty;
	    p1.x=x;
	    p1.y=y;
	    line(p0,p1,m);
	}
	else if(outline->type == T1_PATHTYPE_BEZIER)
	{
	    plotxy p0;
	    plotxy p1;
	    plotxy p2;
	    plotxy p3;
	    T1_BEZIERSEGMENT*o2 = (T1_BEZIERSEGMENT*)outline;
	    p0.x=x; 
	    p0.y=y;
	    p1.x=o2->C.x/(float)0xffff+lastx;
	    p1.y=o2->C.y/(float)0xffff+lasty;
	    p2.x=o2->B.x/(float)0xffff+lastx;
	    p2.y=o2->B.y/(float)0xffff+lasty;
	    p3.x=lastx;
	    p3.y=lasty;
	    spline(p0,p1,p2,p3,m);
	} 
	else {
	 logf("<error> drawpath: unknown outline type:%d\n", outline->type);
	}
	lastx=x;
	lasty=y;
	outline = outline->link;
    }
}

/* process a character. */
void drawchar(struct swfoutput*obj, int t1fontindex, char character, swfmatrix*m)
{

    T1_OUTLINE*outline;
    int width = T1_GetCharWidth(t1fontindex, character);
    BBox bbox = T1_GetCharBBox(t1fontindex, character);
    char*charname= T1_GetCharName(t1fontindex, character);
    logf("<debug> Font name is %s", T1_GetFontFileName(t1fontindex));
    logf("<debug> char 0x%02x is named %s\n",character,charname);
    logf("<debug> bbox: %d %d %d %d\n",bbox.llx,bbox.lly,bbox.urx,bbox.ury);
    if(!charname || charname[0] == '.') 
    {
     /* for newline, we don't print an error. FIXME: We shouldn't get newlines here
	in the first place
      */
     if(character != '\n') {
	 logf("<error> Char to set is not defined!");
	 logf("<error> -  font file is %s\n", T1_GetFontFileName(t1fontindex));
	 logf("<error> -  char 0x%02x is named %s\n",character,charname);
     }
     return;
    }
    swfmatrix m2=*m;    
    m2.m11/=100;
    m2.m21/=100;
    m2.m12/=100;
    m2.m22/=100;
    outline =  T1_GetCharOutline( t1fontindex, character, 100.0, 0);

    if(!lastwasfill)
     ShapeSetStyle(tag,shape,0x8000,fillstyleid,0);
    lastwasfill = 1;

    drawpath(outline, &m2);
}

/* draw a curved polygon. */
void swfoutput_drawpath(swfoutput*output, T1_OUTLINE*outline, struct swfmatrix*m)
{
    if(lastwasfill)
     ShapeSetStyle(tag,shape,linestyleid,0x8000,0);
    lastwasfill = 0;

    drawpath(outline,m); 
}

/* set's the t1 font index of the font to use for swfoutput_drawchar(). */
int swfoutput_setfont(struct swfoutput*obj, int t1id)
{
    obj->t1font = t1id;
}

/* set's the matrix which is to be applied to characters drawn by
   swfoutput_drawchar() */
void swfoutput_setfontmatrix(struct swfoutput*obj,double m11,double m12,
                                                  double m21,double m22)
{
    obj->fontm11 = m11;
    obj->fontm12 = m12;
    obj->fontm21 = m21;
    obj->fontm22 = m22;
}

/* draws a character at x,y. */
void swfoutput_drawchar(struct swfoutput* obj,double x,double y,char character) 
{
    swfmatrix m;
    m.m11 = obj->fontm11;
    m.m12 = obj->fontm12;
    m.m21 = obj->fontm21;
    m.m22 = obj->fontm22;
    m.m13 = x;
    m.m23 = y;
    drawchar(obj, obj->t1font, character, &m);
}

void startshape();
/* initialize the swf writer */
void swfoutput_init(struct swfoutput* obj, char*_filename, int _sizex, int _sizey) 
{
  GLYPH *glyph;
  RGBA rgb;
  SRECT r;
  filename = _filename;
  sizex = _sizex;
  sizey = _sizey;

  logf("<verbose> initializing swf output for size %d*%d\n", sizex,sizey);

  obj->t1font = 0;
  
  memset(&swf,0x00,sizeof(SWF));

  swf.FileVersion    = 4;
//  swf.FrameRate      = 0x1900;
  swf.FrameRate      = 0x0040; // 1 frame per 4 seconds
  swf.MovieSize.xmax = 20*sizex;
  swf.MovieSize.ymax = 20*sizey;
  
  swf.FirstTag = InsertTag(NULL,ST_SETBACKGROUNDCOLOR);
  tag = swf.FirstTag;
  rgb.r = 0xff;
  rgb.g = 0xff;
  rgb.b = 0xff;
  SetRGB(tag,&rgb);
  if(flag_protected)
    tag = InsertTag(tag, ST_PROTECT);
  startshape();
}

void swfoutput_setprotected() //write PROTECT tag
{
  flag_protected = 1;
}

void startshape()
{
  RGBA rgb;
  SRECT r;
  tag = InsertTag(tag,ST_DEFINESHAPE);

  NewShape(&shape);
  rgb.r = 0x00;
  rgb.b = 0x00;
  rgb.g = 0x00;
  linestyleid = ShapeAddLineStyle(shape,1,&rgb);
  fillstyleid = ShapeAddSolidFillStyle(shape,&rgb);

  SetU16(tag,++shapeid);  // ID

  r.xmin = 0;
  r.ymin = 0;
  r.xmax = 20*sizex;
  r.ymax = 20*sizey;
  
  SetRect(tag,&r);

  SetShapeStyles(tag,shape);
  ShapeCountBits(shape,NULL,NULL);
  SetShapeBits(tag,shape);

  ShapeSetAll(tag,shape,/*x*/0,/*y*/0,linestyleid,0,0);
  lastwasfill = 0;
}

void swfoutput_newpage(struct swfoutput*)
{
    ShapeSetEnd(tag);
    tag = InsertTag(tag,ST_PLACEOBJECT2);
    ObjectPlace(tag,shapeid,/*depth*/1,NULL,NULL,NULL);

    tag = InsertTag(tag,ST_SHOWFRAME);
    
    tag = InsertTag(tag,ST_REMOVEOBJECT2);
    SetU16(tag,1);

    startshape();
}

/* "destroy" like in (oo-terminology) "destructor". Perform cleaning
   up, complete the swf, and write it out. */
void swfoutput_destroy(struct swfoutput* obj) 
{
    T1_CloseLib();
    if(!filename) 
	return;
    if(filename)
     fi = open(filename, O_CREAT|O_TRUNC|O_WRONLY, 0777);
    else
     fi = 1; // stdout
    
    if(fi<=0) {
     logf("<fatal> Could not create \"%s\". ", filename);
     exit(1);
    }
  
    ShapeSetEnd(tag);
    tag = InsertTag(tag,ST_PLACEOBJECT2);
	  ObjectPlace(tag,shapeid,1,NULL,NULL,NULL);
    tag = InsertTag(tag,ST_SHOWFRAME);
    tag = InsertTag(tag,ST_END);

    if FAILED(WriteSWF(fi,&swf)) 
     logf("<error> WriteSWF() failed.\n");
    if(filename)
     close(fi);
    printf("SWF written\n");
}

